from collections import deque
from contextlib import contextmanager

from gyver.attrs import mutable

from gyver.exc import FailedFileOperation
from gyver.exc import InvalidPath
from gyver.utils import merge_dicts

from .interface import AbstractFileTree
from .typedef import File
from .typedef import Folder


@mutable
class VirtualFileTree(AbstractFileTree[Folder]):
    base_dir: Folder

    @property
    def root(self):
        return self.base_dir

    @property
    def name(self):
        return self.root.name

    def from_virtual(
        self, virtual_filetree: "VirtualFileTree", *path: str, strict: bool = True
    ):
        target = self.root
        for item in path:
            if item not in target.contents:
                target.contents[item] = Folder.new(item)
            target = target.contents[item]
            if isinstance(target, File):
                raise InvalidPath("Foldername conflicts with file")

        current_value = target.contents.get(virtual_filetree.name)
        if current_value is not None:
            if isinstance(current_value, File):
                raise InvalidPath("Foldername conflicts with file")
            elif isinstance(current_value, Folder) and current_value.contents:
                if not strict:
                    self._try_merge(current_value, virtual_filetree.root)
                else:
                    raise InvalidPath("Non-empty folder with same name already exists")

        target.contents[virtual_filetree.name] = current_value or virtual_filetree.root
        return virtual_filetree.root

    def _try_merge(self, current_target: Folder, virtual_folder: Folder):
        stack = deque([(current_target.contents, virtual_folder.contents)])
        while stack:
            current_contents, virtual_contents = stack.pop()
            for value in current_contents.values():
                if (
                    isinstance(value, Folder)
                    and value.name in virtual_contents
                    and isinstance(virtual_contents[value.name], Folder)
                ):
                    stack.append(
                        (value.contents, virtual_contents[value.name].contents)
                    )
            current_contents.update(
                merge_dicts(current_contents, virtual_contents, on_conflict="left")
            )

    @contextmanager
    def virtual_context(self, dirname: str, *path: str):
        try:
            folder = self.get_dir(dirname, *path) or Folder.new(dirname)
            vt = VirtualFileTree(folder)
            yield vt
        except Exception as e:
            raise FailedFileOperation(
                "unable to complete virtual context after exception", e
            ) from e
        else:
            self.from_virtual(vt, *path, strict=False)



import typing

import orjson

loads: typing.Callable[
    [typing.Union[str, bytes, bytearray, memoryview]], typing.Any
] = orjson.loads
"""
Deserialize a JSON string to a Python object.

Args:
    s (bytes, str, bytearray, memoryview): The JSON string to be deserialized.

Returns:
    typing.Any: The deserialized Python object.
"""


def dumps(val: typing.Any, *, default: typing.Any = None) -> str:
    """
    Serialize a Python object to a JSON-formatted string.

    Args:
        val (typing.Any): The Python object to be serialized.
        default (typing.Any, optional): A default value to be used if an object is not serializable.
            Defaults to None.

    Returns:
        str: The JSON-formatted string representing the serialized object.
    """
    return orjson.dumps(val, default=default).decode()


def load(fdes: typing.TextIO) -> str:
    """
    Deserialize a JSON-formatted string from a text file object.

    Args:
        fdes (typing.TextIO): The text file object to read the JSON-formatted string from.

    Returns:
        typing.Any: The deserialized Python object.
    """
    return loads(fdes.read())


def dump(val: typing.Any, fp: typing.TextIO, *, default: typing.Any = None):
    """
    Serialize a Python object and write the JSON-formatted string to a text file object.

    Args:
        val (typing.Any): The Python object to be serialized.
        fp (typing.TextIO): The text file object to write the JSON-formatted string to.
        default (typing.Any, optional): A default value to be used if an object is not serializable.
            Defaults to None.
    """
    return fp.write(dumps(val, default=default))



import importlib
import inspect
import os
import pathlib
import sys
import threading
import typing

from gyver.attrs import define
from lazyfields import lazyfield
from typing_extensions import Self

from gyver.exc import InvalidPath
from gyver.exc import MissingParams

PathConverter = typing.Callable[[pathlib.Path], str]
StrOrPath = typing.Union[str, pathlib.Path]
T = typing.TypeVar("T")
Validator = typing.Callable[[typing.Any], bool]


def make_modulename_resolver(
    root: pathlib.Path, path_converter: PathConverter = pathlib.Path.as_posix
) -> PathConverter:
    """
    Create a module name resolver based on a root path.

    Args:
        root (pathlib.Path): The root path to resolve from.
        path_converter (PathConverter, optional): A path converter function (default: pathlib.Path.as_posix).

    Returns:
        PathConverter: A callable that resolves a path into a module name.
    """
    root_str = path_converter(root)

    def module_resolver(path: pathlib.Path) -> str:
        target_str = path_converter(path)
        return (
            target_str.replace(root_str, root.name)
            .removesuffix(".py")
            .replace("/", ".")
        )

    return module_resolver


def iterate_module(
    modpath: pathlib.Path, resolver: PathConverter
) -> typing.Generator[tuple[str, typing.Any], None, None]:
    """
    Iterate through a module using a resolver to map paths to module names.

    Args:
        modpath (pathlib.Path): The path to the module.
        resolver (PathConverter): A path converter to map paths to module names.

    Yields:
        tuple[str, Any]: A tuple containing the name and object of each item in the module.
    """
    mod = importlib.import_module(resolver(modpath))
    yield from inspect.getmembers(mod)


@define
class _Finder:
    """
    Finder class to locate and validate objects in a given directory structure.

    Attributes:
        validator (Validator): The validation function for objects.
        root (pathlib.Path): The root path to start searching.
        look_on (Optional[pathlib.Path]): The path to focus the search on (default: None).
        exclude (Sequence[StrOrPath]): Paths to exclude from search (default: []).
    """

    validator: Validator
    root: pathlib.Path
    look_on: typing.Optional[pathlib.Path] = None
    exclude: typing.Sequence[StrOrPath] = ()

    def __iter__(self):
        """
        Create an iterator for the finder.
        """
        return self._iterdir(self.root)

    def find(self):
        """
        Find and return the output dictionary containing the located objects.

        Returns:
            dict: The dictionary containing the located objects.
        """
        return self.output

    @lazyfield
    def output(self):
        return dict(self)

    def __attrs_post_init__(self):
        if not self.root.exists():
            raise InvalidPath(f"root must be a valid path, received {self.root}")

    def _should_look(self, path: pathlib.Path):
        """
        Determine whether the given path should be explored or not.

        Args:
            path (pathlib.Path): The path to evaluate.

        Returns:
            bool: True if the path should be explored, False otherwise.
        """
        str_exclude = tuple(item for item in self.exclude if isinstance(item, str))
        path_exclude = tuple(
            item for item in self.exclude if isinstance(item, pathlib.Path)
        )
        return (
            path.name not in str_exclude
            and path not in path_exclude
            and (not path.is_file() or path.name.endswith(".py"))
        )

    @lazyfield
    def target_path(self):
        """
        Get the target path for exploration.
        """
        return self.look_on or self.root

    @lazyfield
    def resolver(self) -> PathConverter:
        """
        Get the path converter based on the root path.

        Returns:
            PathConverter: The path converter function.
        """
        return make_modulename_resolver(self.root)

    def _itermod(self, path: pathlib.Path):
        """
        Iterate through a module using the given resolver.

        Args:
            path (pathlib.Path): The path to the module.

        Yields:
            tuple[str, Any]: A tuple containing the name and object of each item in the module.
        """
        yield from (
            (name, obj)
            for name, obj in iterate_module(path, self.resolver)
            if self.validator(obj)
        )

    def _iterdir(
        self, path: pathlib.Path
    ) -> typing.Generator[tuple[str, typing.Any], None, None]:
        """
        Iterate through a directory, exploring subdirectories and modules.

        Args:
            path (pathlib.Path): The path to the directory.

        Yields:
            tuple[str, Any]: A tuple containing the name and object of each item in the directory.
        """
        for item in path.iterdir():
            if not self._should_look(item):
                continue
            if item.is_dir():
                yield from self._iterdir(item)
            else:
                yield from self._itermod(item)


def class_validator(*types: type) -> Validator:
    """
    Create a class validator that checks if objects are subclasses of the provided types.

    Args:
        *types (type): Types to validate against.

    Returns:
        Validator: The validator function.
    """
    if not types:
        raise MissingParams("Missing types in class_validator paramaters")

    def validator(obj: typing.Any):
        return (
            issubclass(obj, types) and obj not in types
            if inspect.isclass(obj)
            else False
        )

    return validator


def instance_validator(*types: type) -> Validator:
    """
    Create an instance validator that checks if objects are instances of the provided types.

    Args:
        *types (type): Types to validate against.

    Returns:
        Validator: The validator function.
    """
    if not types:
        raise MissingParams("Missing types in instance_validator paramaters")

    def validator(obj: typing.Any):
        return isinstance(obj, types)

    return validator


class FinderBuilder:
    """
    Builder class to construct a Finder instance with various configurations.
    """

    def __init__(self) -> None:
        self.root: typing.Optional[pathlib.Path] = None
        self._validators = []
        self._lock = threading.Lock()
        self._exclude = []

    def reset(self) -> None:
        """
        Reset the builder to its initial state.
        """
        with self._lock:
            self.root = None
            self._validators = []
            self._exclude = []

    def from_path(self, path: pathlib.Path) -> Self:
        """
        Set the root path for the finder from a provided path.

        Args:
            path (pathlib.Path): The root path.

        Returns:
            FinderBuilder: The FinderBuilder instance.

        Raises:
            InvalidPath: If the path does not exist
        """
        if not path.exists():
            raise InvalidPath(f"root must be a valid path, received {path}")
        self.root = path
        return self

    def from_cwd(self) -> Self:
        """
        Set the root path for the finder to the current working directory.

        Returns:
            FinderBuilder: The FinderBuilder instance.
        """
        return self.from_path(pathlib.Path.cwd())

    def from_project(self) -> Self:
        """
        Set the root path for the finder to the project root.

        Returns:
            FinderBuilder: The FinderBuilder instance.
        """
        frame_stack: list[inspect.FrameInfo] = inspect.stack()

        frame_info: inspect.FrameInfo = frame_stack[1]

        if frame_info.filename == __file__:
            frame_info = next(
                frame for frame in frame_stack if frame.filename != __file__
            )

        caller_path: str = frame_info.filename
        caller_absolute_path: str = os.path.abspath(caller_path)
        caller_root_path = next(
            iter(
                sorted(
                    (p for p in sys.path if p in caller_absolute_path),
                    key=lambda p: len(p),
                )
            )
        )

        if not os.path.isabs(caller_path):
            caller_module_name: str = pathlib.Path(caller_path).name
            project_related_folders: str = caller_path.replace(
                os.sep + caller_module_name, ""
            )

            caller_root_path = caller_root_path.replace(project_related_folders, "")

        return self.from_path(pathlib.Path(caller_root_path))

    def from_package(self, stack_position: int = 1) -> Self:
        """
        Set the root path for the finder to the calling package root.

        Args:
            stack_position (int, optional): The position in the stack trace of the calling function (default: 1).

        Returns:
            FinderBuilder: The FinderBuilder instance.
        """

        frame_stack = inspect.stack()

        frame_info = frame_stack[stack_position]

        if frame_info.filename == __file__:
            frame_info = next(
                frame for frame in frame_stack if frame.filename != __file__
            )

        package: str = frame_info.frame.f_globals["__name__"]
        steps_to_package = len(package.split("."))
        caller_root_path = pathlib.Path(package.replace(".", "/")).resolve()
        for _ in range(steps_to_package - 1):
            caller_root_path = caller_root_path.parent
        return self.from_path(pathlib.Path(caller_root_path))

    def _make_validator(self) -> Validator:
        validators = [*self._validators]

        def _validator(obj: typing.Any):
            return all(validator(obj) for validator in validators)

        return _validator

    def add_validator(self, validator: Validator) -> Self:
        """
        Add a validation function to the builder.

        Args:
            validator (Validator): The validation function to add.

        Returns:
            FinderBuilder: The FinderBuilder instance.
        """
        with self._lock:
            self._validators.append(validator)
        return self

    def child_of(self, *types: type) -> Self:
        """
        Add a class validator that checks if objects are subclasses of the provided types.

        Args:
            *types (type): Types to validate against.

        Returns:
            FinderBuilder: The FinderBuilder instance.
        """
        return self.add_validator(class_validator(*types))

    def instance_of(self, *types: type) -> Self:
        """
        Add an instance validator that checks if objects are instances of the provided types.

        Args:
            *types (type): Types to validate against.

        Returns:
            FinderBuilder: The FinderBuilder instance.
        """
        return self.add_validator(instance_validator(*types))

    def exclude(self, *paths: StrOrPath) -> Self:
        """
        Exclude paths from search.

        Args:
            *paths (StrOrPath): Paths to exclude.

        Returns:
            FinderBuilder: The FinderBuilder instance.
        """
        self._exclude.extend(paths)
        return self

    def exclude_default(self) -> Self:
        """
        Exclude default paths (e.g., '__init__.py', '__pycache__') from search.

        Returns:
            FinderBuilder: The FinderBuilder instance.
        """
        return self.exclude("__init__.py", "__pycache__")

    def build(
        self,
        look_on: typing.Optional[pathlib.Path] = None,
    ) -> _Finder:
        """
        Build a Finder instance with the provided configurations.

        Args:
            look_on (Optional[pathlib.Path], optional): The path to focus the search on (default: None).

        Returns:
            _Finder: The built Finder instance.
        """
        with self._lock:
            if self.root is None:
                self.from_package()
            finder = _Finder(
                self._make_validator(),
                typing.cast(pathlib.Path, self.root),
                look_on,
            )
        self.reset()
        return finder

    def find(
        self,
        look_on: typing.Optional[pathlib.Path] = None,
    ) -> dict[str, typing.Any]:
        """
        Find and return located objects using the provided configurations.

        Args:
            look_on (Optional[pathlib.Path], optional): The path to focus the search on (default: None).

        Returns:
            dict: The dictionary containing the located objects.
        """
        if self.root is None:
            self.from_project()
            assert self.root is not None
        return self.build(look_on).find()


finder_builder = FinderBuilder()



import inspect
import warnings
from collections.abc import Callable
from functools import WRAPPER_UPDATES, update_wrapper
from inspect import signature
from typing import Any, Coroutine, Literal, TypeVar

from config import MISSING
from config.interface import ConfigLike
from lazyfields import asynclazyfield, lazyfield
from typing_extensions import Concatenate, ParamSpec

from gyver.attrs import call_init, define

from .factory import _DEFAULT_CONFIG, AdapterConfigFactory
from .mark import is_config

T = TypeVar("T")
SelfT = TypeVar("SelfT")
P = ParamSpec("P")


def load(
    cls: type[T],
    __prefix__: str = "",
    __sep__: str = "__",
    __kwargs_type__: Literal["presets", "defaults"] = "defaults",
    __config__: ConfigLike = _DEFAULT_CONFIG,
    **kwargs: Any,
) -> T:
    """
    Load configuration for a given class.

    Args:
        cls (type): The class to load configuration for.
        __prefix__ (str, optional): Prefix to prepend to configuration keys. Defaults to "".
        __sep__ (str, optional): Separator used between prefix and configuration keys. Defaults to "__".
        __kwargs_type__ (Literal["presets", "defaults"], optional):
            Type of configuration to load, either "presets" or "defaults". Defaults to "defaults".
        __config__ (ConfigLike, optional): Configuration backend to use. Defaults to _DEFAULT_CONFIG.
        **kwargs (Any): Additional keyword arguments.

    Returns:
        T: An instance of the class with configuration loaded.

    Raises:
        (Any): Raises exceptions related to loading and casting configuration values.
    """
    factory = AdapterConfigFactory(__config__)
    defaults = kwargs if __kwargs_type__ == "defaults" else {}
    return factory.load(
        cls,
        __prefix__=__prefix__,
        __sep__=__sep__,
        __kwargs_type__=__kwargs_type__,
        presets=kwargs if __kwargs_type__ == "presets" else None,
        **defaults,
    )


def parametrize(
    call: Callable[..., T],
    __prefix__: str = "",
    __sep__: str = "__",
    __kwargs_type__: Literal["presets", "defaults"] = "defaults",
    __config__: ConfigLike = _DEFAULT_CONFIG,
    __transform__: Callable[[str], str] = str.upper,
    **kwargs: Any,
) -> T:
    """
    Parameterize a callable function with configuration values.
    It is recommended to store the result of this function in a constant to avoid recomputation.

    Args:
        call (Callable): The callable function to parameterize.
        __prefix__ (str, optional): Prefix to prepend to parameter names. Defaults to "".
        __sep__ (str, optional): Separator used between prefix and parameter names. Defaults to "__".
        __kwargs_type__ (Literal["presets", "defaults"], optional):
            Type of configuration to load, either "presets" or "defaults". Defaults to "defaults".
        __config__ (ConfigLike, optional): Configuration backend to use. Defaults to _DEFAULT_CONFIG.
        __transform__ ((str) -> str): A callable that transforms the parameter names when searching the environment for configuration values.
            The default transformation is set to `str.upper`, converting names to uppercase.
        **kwargs (Any): Additional keyword arguments.

    Returns:
        T: The result of the callable function with parameterized values.

    Raises:
        (Any): Raises exceptions related to loading and casting configuration values.
    """
    sig = signature(call)
    prefix = (
        f"{__prefix__}_" if __prefix__ and not __prefix__.endswith("_") else __prefix__
    )
    args = {
        key: (
            __transform__(f"{prefix}{key}"),
            str if value.annotation is inspect._empty else value.annotation,
        )
        for key, value in sig.parameters.items()
        if not (key in kwargs and __kwargs_type__ == "presets")
    }
    defaults = (
        {__transform__(key): value for key, value in kwargs.items()} | kwargs
        if __kwargs_type__ == "defaults"
        else {}
    )
    presets = kwargs if __kwargs_type__ == "presets" else {}
    if not args:
        warnings.warn(
            "No args could be inspected in function: {}".format(call.__name__)
        )
    vals = {
        key: __config__(prefix, _cast, defaults.get(key, MISSING))
        if not is_config(_cast)
        else load(
            _cast,
            __prefix__=prefix,
            __sep__=__sep__,
            __config__=__config__,
            __kwargs_type__=__kwargs_type__,
            **{key: value for key, value in defaults.items() if key.startswith(prefix)},
        )
        for key, (prefix, _cast) in args.items()
    } | presets
    return call(**vals)


@define
class AttributeLoader:
    __prefix__: str = ""
    __sep__: str = "__"
    __config__: ConfigLike = _DEFAULT_CONFIG
    __self_name__: str = "self"
    __transform__: Callable[[str], str] = str.upper

    def __init__(
        self,
        __prefix__: str = "",
        __sep__: str = "__",
        __config__: ConfigLike = _DEFAULT_CONFIG,
        __self_name__: str = "self",
        __transform__: Callable[[str], str] = str.upper,
    ) -> None:
        call_init(self, __prefix__, __sep__, __config__, __self_name__, __transform__)

    def lazy(self, method: Callable[Concatenate[SelfT, P], T]) -> lazyfield[SelfT, T]:
        def _parametrize(__self__: SelfT) -> T:
            return parametrize(
                method,
                __prefix__=self.__prefix__,
                __sep__=self.__sep__,
                __config__=self.__config__,
                __kwargs_type__="presets",
                __transform__=self.__transform__,
                **{self.__self_name__: __self__},
            )

        update_wrapper(
            _parametrize,
            method,
            ("__module__", "__name__", "__qualname__", "__doc__"),
            updated=WRAPPER_UPDATES,
        )
        return lazyfield(_parametrize)

    def asynclazy(
        self, method: Callable[Concatenate[SelfT, P], Coroutine[Any, Any, T]]
    ) -> asynclazyfield[SelfT, T]:
        def _parametrize(__self__: SelfT) -> Coroutine[Any, Any, T]:
            return parametrize(
                method,
                __prefix__=self.__prefix__,
                __sep__=self.__sep__,
                __config__=self.__config__,
                __kwargs_type__="presets",
                __transform__=self.__transform__,
                **{self.__self_name__: __self__},
            )

        update_wrapper(
            _parametrize,
            method,
            ("__module__", "__name__", "__qualname__", "__doc__"),
            updated=WRAPPER_UPDATES,
        )
        return asynclazyfield(_parametrize)


attribute = AttributeLoader()



from typing import TypeVar

T = TypeVar("T", bound=Exception)


def panic(exc: type[T], message: str, *args) -> T:
    """
    Create an instance of the specified exception class with a modified error message.

    This function creates an exception instance by calling the constructor of the
    specified exception class with the modified error message and any additional
    arguments provided.

    Args:
        exc (type[T]): The exception class to instantiate.
        message (str): The error message for the exception.
        *args (Any): Additional arguments to pass to the exception constructor.

    Returns:
        T: An instance of the specified exception class.

    Examples:
        >>> class CustomError(Exception):
        ...     pass
        ...
        >>> error = panic(CustomError, "Something went wrong", "additional_info")
        >>> isinstance(error, CustomError)
        True
        >>> str(error)
        'Something went wrong! additional_info'
    """
    return exc(message.removesuffix("!") + "!", *args)



from datetime import date
from datetime import datetime
from datetime import timezone
from datetime import tzinfo
from typing import Callable


def now(tz: tzinfo = timezone.utc) -> datetime:
    """
    Get the current aware datetime.

    Args:
        tz (tzinfo): The timezone for the datetime (default is UTC).

    Returns:
        datetime: The current datetime in the specified timezone.
    """
    return datetime.now(tz)


def today(tz: tzinfo = timezone.utc) -> date:
    """
    Get today's aware date.

    Args:
        tz (tzinfo): The timezone for the date (default is UTC).

    Returns:
        date: Today's date in the specified timezone.
    """
    return now(tz).date()


def make_now_factory(tz: tzinfo) -> Callable[[], datetime]:
    """
    Create a function to get the current aware datetime with a specific timezone.

    Args:
        tz (tzinfo): The timezone for the datetime factory.

    Returns:
        Callable[[], datetime]: A function that returns the current datetime in the specified timezone.
    """

    def _now():
        return now(tz)

    return _now



import functools
import warnings
from collections import deque
from collections.abc import Callable
from itertools import chain
from typing import TYPE_CHECKING, Literal, TypeVar, cast
from weakref import WeakValueDictionary

from typing_extensions import ParamSpec

from gyver.attrs import define, info
from gyver.exc import MergeConflict

P = ParamSpec("P")
T = TypeVar("T")


def cache(f: Callable[P, T]) -> Callable[P, T]:
    """
    Cache the result of a function.
    It uses `functools.cache` and just
    adds proper type hints to it.

    Args:
        f (Callable[P, T]): The function to be cached.

    Returns:
        Callable[P, T]: The cached version of the function.
    """
    return cast(Callable[P, T], functools.cache(f))


def deprecated(func: Callable[P, T]) -> Callable[P, T]:
    """
    Mark a function as deprecated and issue a warning when it's used.

    Args:
        func (Callable[P, T]): The function to be marked as deprecated.

    Returns:
        Callable[P, T]: A wrapped version of the function that issues a warning on use.
    """

    @functools.wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> T:
        if not hasattr(func, "__warn_deprecated__"):
            warnings.warn(
                f"{func.__qualname__} is "
                "deprecated and can be removed without notice",
                DeprecationWarning,
            )
            func.__warn_deprecated__ = True
        return func(*args, **kwargs)

    return inner


if TYPE_CHECKING:
    import typing_extensions as te

    deprecated = te.deprecated  # type: ignore


class DeprecatedClass:
    """
    A class to mark as deprecated.

    This class issues a deprecation warning when instantiated.
    """

    __warn_deprecated__ = False

    def __init__(self) -> None:
        if not type(self).__warn_deprecated__:
            warnings.warn(
                f"{'.'.join((type(self).__module__, type(self).__name__))} is "
                "deprecated and can be removed without notice",
                DeprecationWarning,
            )
            type(self).__warn_deprecated__ = True


def merge_dicts(
    left: dict,
    right: dict,
    on_conflict: Literal["strict", "left", "right"],
    merge_sequences: bool = True,
) -> dict:
    """
    Merge two dictionaries with customizable conflict resolution strategy.

    Args:
        left (dict): The left dictionary to merge.
        right (dict): The right dictionary to merge.
        on_conflict (Literal["strict", "left", "right"]): The conflict resolution strategy to use.

            - 'strict': Raise a MergeConflict exception if conflicts occur.
            - 'left': Prioritize the values from the left dictionary in case of conflicts.
            - 'right': Prioritize the values from the right dictionary in case of conflicts.
        merge_sequences (bool, optional): Indicates whether to merge sequences (lists, sets, tuples) or skip them.

            - If True, sequences will be merged based on the conflict resolution strategy.
            - If False, sequences will be skipped, and the value from the chosen (defaults to left on strict)
            dictionary will be used. Default is True.

    Returns:
        dict: The merged dictionary.

    Raises:
        MergeConflict: If conflicts occur and the conflict resolution strategy is set to 'strict'.
    """

    output = {key: value for key, value in left.items() if key not in right}

    stack = deque([(left, right, output)])

    while stack:
        left_curr, right_curr, output_curr = stack.pop()

        for key, value in right_curr.items():
            if key not in left_curr:
                output_curr[key] = value
            elif isinstance(value, (list, set, tuple)) and merge_sequences:
                left_val = left_curr[key]
                if isinstance(left_val, (list, set, tuple)):
                    type_ = type(value) if on_conflict == "right" else type(left_val)
                    output_curr[key] = type_(chain(left_val, value))
            elif isinstance(value, dict):
                if isinstance(left_curr[key], dict):
                    output_curr[key] = {
                        lkey: lvalue
                        for lkey, lvalue in left_curr[key].items()
                        if lkey not in value
                    }
                    stack.append((left_curr[key], value, output_curr[key]))
            elif on_conflict not in ("left", "right"):
                raise MergeConflict(
                    "Conflict found when trying to merge dicts",
                    key,
                    value,
                    left_curr[key],
                )
            elif on_conflict == "left":
                output_curr[key] = left_curr[key]
            else:
                output_curr[key] = value

    return output


@define
class Cached:
    _mapping: WeakValueDictionary = info(default_factory=WeakValueDictionary)
